/******************************************************************************
 * Copyright (c) 2019 Philipp Schubert.
 * All rights reserved. This program and the accompanying materials are made
 * available under the terms of:
 * https://github.com/secure-software-engineering/phasar/blob/master/LICENSE.txt
 *
 * Contributors:
 *     Philipp Schubert and others
 *****************************************************************************/

#include <iostream>
#include <memory>

#include <boost/filesystem/operations.hpp>

#include <llvm/IR/Instruction.h>
#include <llvm/IR/Instructions.h>

#include <phasar/DB/ProjectIRDB.h>
#include <phasar/PhasarLLVM/ControlFlow/LLVMBasedICFG.h>
#include <phasar/PhasarLLVM/IfdsIde/FlowFunctions/Identity.h>
#include <phasar/PhasarLLVM/IfdsIde/LLVMDefaultIFDSTabulationProblem.h>
#include <phasar/PhasarLLVM/IfdsIde/LLVMZeroValue.h>
#include <phasar/PhasarLLVM/IfdsIde/Solver/LLVMIFDSSolver.h>
#include <phasar/PhasarLLVM/Pointer/LLVMTypeHierarchy.h>
#include <phasar/PhasarLLVM/Utils/TaintSensitiveFunctions.h>
#include <phasar/Utils/LLVMIRToSrc.h>
#include <phasar/Utils/LLVMShorthands.h>
#include <phasar/Utils/Logger.h>

using namespace std;
using namespace psr;

class TaintAnalysisProblem
    : public LLVMDefaultIFDSTabulationProblem<const llvm::Value *,
                                              LLVMBasedICFG &> {
 private:
  TaintSensitiveFunctions SourceSinkFunctions;
  std::vector<std::string> EntryPoints;

 public:
  std::map<const llvm::Instruction *, std::set<const llvm::Value *>> Leaks;

  TaintAnalysisProblem(LLVMBasedICFG &icfg, const LLVMTypeHierarchy &th,
                       const ProjectIRDB &irdb, TaintSensitiveFunctions TSF,
                       std::vector<std::string> EntryPoints = {"main"})
      : LLVMDefaultIFDSTabulationProblem(icfg, th, irdb),
        SourceSinkFunctions(TSF),
        EntryPoints(EntryPoints) {
    TaintAnalysisProblem::zerovalue = createZeroValue();
  }

  virtual ~TaintAnalysisProblem() = default;

  std::shared_ptr<FlowFunction<const llvm::Value *>> getNormalFlowFunction(
      const llvm::Instruction *curr, const llvm::Instruction *succ) override {
    // TODO: implement
    return Identity<const llvm::Value *>::getInstance();
  }

  std::shared_ptr<FlowFunction<const llvm::Value *>> getCallFlowFunction(
      const llvm::Instruction *callStmt,
      const llvm::Function *destMthd) override {
    // TODO: implement
    return Identity<const llvm::Value *>::getInstance();
  }

  std::shared_ptr<FlowFunction<const llvm::Value *>> getRetFlowFunction(
      const llvm::Instruction *callSite, const llvm::Function *calleeMthd,
      const llvm::Instruction *exitStmt,
      const llvm::Instruction *retSite) override {
    // TODO: implement
    return Identity<const llvm::Value *>::getInstance();
  }

  std::shared_ptr<FlowFunction<const llvm::Value *>> getCallToRetFlowFunction(
      const llvm::Instruction *callSite, const llvm::Instruction *retSite,
      std::set<const llvm::Function *> callees) override {
    // TODO: implement
    return Identity<const llvm::Value *>::getInstance();
  }

  std::shared_ptr<FlowFunction<const llvm::Value *>> getSummaryFlowFunction(
      const llvm::Instruction *callStmt,
      const llvm::Function *destMthd) override {
    return nullptr;
  }

  std::map<const llvm::Instruction *, std::set<const llvm::Value *>>
  initialSeeds() override {
    auto &lg = lg::get();
    LOG_IF_ENABLE(BOOST_LOG_SEV(lg, DEBUG)
                  << "IFDSTaintAnalysis::initialSeeds()");
    // If main function is the entry point, commandline arguments have to be
    // tainted. Otherwise, we just use the zero value to initialize the
    // analysis.
    map<const llvm::Instruction *, set<const llvm::Value *>> SeedMap;
    for (auto &EntryPoint : EntryPoints) {
      if (EntryPoint == "main") {
        set<const llvm::Value *> CmdArgs;
        for (auto &Arg : icfg.getMethod(EntryPoint)->args()) {
          CmdArgs.insert(&Arg);
        }
        CmdArgs.insert(zeroValue());
        SeedMap.insert(
            make_pair(&icfg.getMethod(EntryPoint)->front().front(), CmdArgs));
      } else {
        SeedMap.insert(make_pair(&icfg.getMethod(EntryPoint)->front().front(),
                                 set<const llvm::Value *>({zeroValue()})));
      }
    }
    return SeedMap;
  }

  const llvm::Value *createZeroValue() override {
    auto &lg = lg::get();
    LOG_IF_ENABLE(BOOST_LOG_SEV(lg, DEBUG)
                  << "IFDSTaintAnalysis::createZeroValue()");
    // create a special value to represent the zero value!
    return LLVMZeroValue::getInstance();
  }

  bool isZeroValue(const llvm::Value *d) const override {
    return isLLVMZeroValue(d);
  }

  void printNode(std::ostream &os, const llvm::Instruction *n) const override {
    os << llvmIRToString(n);
  }

  void printDataFlowFact(std::ostream &os,
                         const llvm::Value *d) const override {
    os << llvmIRToString(d);
  }

  void printMethod(std::ostream &os, const llvm::Function *m) const override {
    os << m->getName().str();
  }

  void printIFDSReport(
      std::ostream &os,
      SolverResults<const llvm::Instruction *, const llvm::Value *,
                    BinaryDomain> &SR) override {
    os << "\n----- Found the following leaks -----\n";
    if (Leaks.empty()) {
      os << "No leaks found!\n";
    } else {
      for (auto Leak : Leaks) {
        os << "At instruction\nIR  : " << llvmIRToString(Leak.first) << '\n';
        os << llvmValueToSrc(Leak.first);
        os << "\n\nLeak(s):\n";
        for (auto LeakedValue : Leak.second) {
          os << "IR  : ";
          // Get the actual leaked alloca instruction if possible
          if (auto Load = llvm::dyn_cast<llvm::LoadInst>(LeakedValue)) {
            os << llvmIRToString(Load->getPointerOperand()) << '\n'
               << llvmValueToSrc(Load->getPointerOperand()) << '\n';

          } else {
            os << llvmIRToString(LeakedValue) << '\n'
               << llvmValueToSrc(LeakedValue) << '\n';
          }
        }
        os << "-------------------\n";
      }
    }
  }
};

int main(int argc, char **argv) {
  initializeLogger(false);
  auto &lg = lg::get();
  if (argc < 2 || !bfs::exists(argv[1]) || bfs::is_directory(argv[1])) {
    std::cerr << "usage: <prog> <ir file>\n";
    std::cerr << "use programs in test/\n";
    return 1;
  }
  initializeLogger(false);
  ProjectIRDB DB({argv[1]}, IRDBOptions::WPA);
  DB.preprocessIR();
  const llvm::Function *F = DB.getFunction("main");
  if (!F) {
    std::cerr << "program does not contain a 'main' function!\n";
    return 1;
  }
  LLVMTypeHierarchy H(DB);
  LLVMBasedICFG ICFG(H, DB, CallGraphAnalysisType::CHA, {"main"});
  TaintSensitiveFunctions TSF;
  TaintAnalysisProblem TA(ICFG, H, DB, TSF);
  LLVMIFDSSolver<const llvm::Value *, LLVMBasedICFG &> Solver(TA, true, true);
  cout << "Hello World!" << endl;
  Solver.solve();
  return 0;
}
